import math
import re
from collections import namedtuple
from functools import lru_cache

import numpy as np

from m2cgen.ast import TOTAL_NUMBER_OF_EXPRESSIONS

CachedResult = namedtuple('CachedResult', ['var_name', 'expr_result'])


def get_file_content(path):
    return path.read_text(encoding="utf-8")


@lru_cache(maxsize=1 << math.ceil(math.log(TOTAL_NUMBER_OF_EXPRESSIONS, 2)))
def _get_handler_name(expr_tpe):
    return f"interpret_{_normalize_expr_name(expr_tpe.__name__)}"


def _normalize_expr_name(name):
    return re.sub("(?!^)([A-Z]+)", r"_\1", name).lower()


def format_float(value):
    return np.format_float_positional(value, unique=True, trim="0")



from pathlib import Path

from m2cgen.exporters import (
    export_to_c,
    export_to_c_sharp,
    export_to_dart,
    export_to_elixir,
    export_to_f_sharp,
    export_to_go,
    export_to_haskell,
    export_to_java,
    export_to_javascript,
    export_to_php,
    export_to_powershell,
    export_to_python,
    export_to_r,
    export_to_ruby,
    export_to_rust,
    export_to_visual_basic
)

__all__ = [
    export_to_c,
    export_to_go,
    export_to_java,
    export_to_python,
    export_to_javascript,
    export_to_visual_basic,
    export_to_c_sharp,
    export_to_powershell,
    export_to_r,
    export_to_php,
    export_to_dart,
    export_to_haskell,
    export_to_ruby,
    export_to_f_sharp,
    export_to_rust,
    export_to_elixir,
]

__version__ = (Path(__file__).absolute().parent / "VERSION.txt").read_text(encoding="utf-8").strip()



from m2cgen.assemblers.boosting import (
    LightGBMModelAssembler,
    XGBoostLinearModelAssembler,
    XGBoostModelAssemblerSelector,
    XGBoostTreeModelAssembler
)
from m2cgen.assemblers.ensemble import RandomForestModelAssembler
from m2cgen.assemblers.linear import (
    ProcessMLEModelAssembler,
    SklearnGLMModelAssembler,
    SklearnLinearModelAssembler,
    StatsmodelsGLMModelAssembler,
    StatsmodelsLinearModelAssembler,
    StatsmodelsModelAssemblerSelector
)
from m2cgen.assemblers.meta import RANSACModelAssembler
from m2cgen.assemblers.svm import LightningSVMModelAssembler, SklearnSVMModelAssembler
from m2cgen.assemblers.tree import TreeModelAssembler

__all__ = [
    SklearnLinearModelAssembler,
    StatsmodelsLinearModelAssembler,
    ProcessMLEModelAssembler,
    RANSACModelAssembler,
    TreeModelAssembler,
    RandomForestModelAssembler,
    XGBoostModelAssemblerSelector,
    XGBoostTreeModelAssembler,
    XGBoostLinearModelAssembler,
    LightGBMModelAssembler,
    SklearnSVMModelAssembler,
    LightningSVMModelAssembler,
    StatsmodelsGLMModelAssembler,
    StatsmodelsModelAssemblerSelector,
    SklearnGLMModelAssembler,
]


SUPPORTED_MODELS = {
    # LightGBM
    "lightgbm_LGBMClassifier": LightGBMModelAssembler,
    "lightgbm_LGBMRegressor": LightGBMModelAssembler,

    # XGBoost
    "xgboost_XGBClassifier": XGBoostModelAssemblerSelector,
    "xgboost_XGBRFClassifier": XGBoostModelAssemblerSelector,
    "xgboost_XGBRegressor": XGBoostModelAssemblerSelector,
    "xgboost_XGBRFRegressor": XGBoostModelAssemblerSelector,

    # Sklearn SVM
    "sklearn_LinearSVC": SklearnLinearModelAssembler,
    "sklearn_LinearSVR": SklearnLinearModelAssembler,
    "sklearn_NuSVC": SklearnSVMModelAssembler,
    "sklearn_NuSVR": SklearnSVMModelAssembler,
    "sklearn_OneClassSVM": SklearnSVMModelAssembler,
    "sklearn_SVC": SklearnSVMModelAssembler,
    "sklearn_SVR": SklearnSVMModelAssembler,

    # Lightning SVM
    "lightning_KernelSVC": LightningSVMModelAssembler,
    "lightning_LinearSVC": SklearnLinearModelAssembler,
    "lightning_LinearSVR": SklearnLinearModelAssembler,

    # Sklearn Linear Regressors
    "sklearn_ARDRegression": SklearnLinearModelAssembler,
    "sklearn_BayesianRidge": SklearnLinearModelAssembler,
    "sklearn_ElasticNet": SklearnLinearModelAssembler,
    "sklearn_ElasticNetCV": SklearnLinearModelAssembler,
    "sklearn_GammaRegressor": SklearnGLMModelAssembler,
    "sklearn_HuberRegressor": SklearnLinearModelAssembler,
    "sklearn_Lars": SklearnLinearModelAssembler,
    "sklearn_LarsCV": SklearnLinearModelAssembler,
    "sklearn_Lasso": SklearnLinearModelAssembler,
    "sklearn_LassoCV": SklearnLinearModelAssembler,
    "sklearn_LassoLars": SklearnLinearModelAssembler,
    "sklearn_LassoLarsCV": SklearnLinearModelAssembler,
    "sklearn_LassoLarsIC": SklearnLinearModelAssembler,
    "sklearn_LinearRegression": SklearnLinearModelAssembler,
    "sklearn_OrthogonalMatchingPursuit": SklearnLinearModelAssembler,
    "sklearn_OrthogonalMatchingPursuitCV": SklearnLinearModelAssembler,
    "sklearn_PassiveAggressiveRegressor": SklearnLinearModelAssembler,
    "sklearn_PoissonRegressor": SklearnGLMModelAssembler,
    "sklearn_RANSACRegressor": RANSACModelAssembler,
    "sklearn_Ridge": SklearnLinearModelAssembler,
    "sklearn_RidgeCV": SklearnLinearModelAssembler,
    "sklearn_SGDRegressor": SklearnLinearModelAssembler,
    "sklearn_TheilSenRegressor": SklearnLinearModelAssembler,
    "sklearn_TweedieRegressor": SklearnGLMModelAssembler,

    # Statsmodels Linear Regressors
    "statsmodels_GLMResultsWrapper": StatsmodelsGLMModelAssembler,
    "statsmodels_ProcessMLEResults": ProcessMLEModelAssembler,
    "statsmodels_RegressionResultsWrapper": StatsmodelsLinearModelAssembler,
    "statsmodels_RegularizedResultsWrapper": StatsmodelsModelAssemblerSelector,

    # Lightning Linear Regressors
    "lightning_AdaGradRegressor": SklearnLinearModelAssembler,
    "lightning_CDRegressor": SklearnLinearModelAssembler,
    "lightning_FistaRegressor": SklearnLinearModelAssembler,
    "lightning_SAGARegressor": SklearnLinearModelAssembler,
    "lightning_SAGRegressor": SklearnLinearModelAssembler,
    "lightning_SDCARegressor": SklearnLinearModelAssembler,
    "lightning_SGDRegressor": SklearnLinearModelAssembler,

    # Sklearn Linear Classifiers
    "sklearn_LogisticRegression": SklearnLinearModelAssembler,
    "sklearn_LogisticRegressionCV": SklearnLinearModelAssembler,
    "sklearn_PassiveAggressiveClassifier": SklearnLinearModelAssembler,
    "sklearn_Perceptron": SklearnLinearModelAssembler,
    "sklearn_RidgeClassifier": SklearnLinearModelAssembler,
    "sklearn_RidgeClassifierCV": SklearnLinearModelAssembler,
    "sklearn_SGDClassifier": SklearnLinearModelAssembler,

    # Lightning Linear Classifiers
    "lightning_AdaGradClassifier": SklearnLinearModelAssembler,
    "lightning_CDClassifier": SklearnLinearModelAssembler,
    "lightning_FistaClassifier": SklearnLinearModelAssembler,
    "lightning_SAGAClassifier": SklearnLinearModelAssembler,
    "lightning_SAGClassifier": SklearnLinearModelAssembler,
    "lightning_SDCAClassifier": SklearnLinearModelAssembler,
    "lightning_SGDClassifier": SklearnLinearModelAssembler,

    # Decision trees
    "sklearn_DecisionTreeClassifier": TreeModelAssembler,
    "sklearn_DecisionTreeRegressor": TreeModelAssembler,
    "sklearn_ExtraTreeClassifier": TreeModelAssembler,
    "sklearn_ExtraTreeRegressor": TreeModelAssembler,

    # Ensembles
    "sklearn_ExtraTreesClassifier": RandomForestModelAssembler,
    "sklearn_ExtraTreesRegressor": RandomForestModelAssembler,
    "sklearn_RandomForestClassifier": RandomForestModelAssembler,
    "sklearn_RandomForestRegressor": RandomForestModelAssembler,
}


def _get_full_model_name(model):
    type_name = type(model)
    return f"{type_name.__module__.split('.')[0]}_{type_name.__name__}"


def get_assembler_cls(model):
    model_name = _get_full_model_name(model)
    assembler_cls = SUPPORTED_MODELS.get(model_name)

    if not assembler_cls:
        raise NotImplementedError(f"Model '{model_name}' is not supported")

    return assembler_cls



from m2cgen import assemblers


class BaseMetaAssembler(assemblers.base.ModelAssembler):

    def assemble(self):
        base_model = self._get_base_model()
        return assemblers.get_assembler_cls(base_model)(base_model).assemble()

    def _get_base_model(self):
        raise NotImplementedError


class RANSACModelAssembler(BaseMetaAssembler):

    def _get_base_model(self):
        return self.model.estimator_



import numpy as np

from m2cgen import ast


def mul(left, right, to_reuse=False):
    return ast.BinNumExpr(left, right, ast.BinNumOpType.MUL, to_reuse=to_reuse)


def div(left, right, to_reuse=False):
    return ast.BinNumExpr(left, right, ast.BinNumOpType.DIV, to_reuse=to_reuse)


def add(left, right, to_reuse=False):
    return ast.BinNumExpr(left, right, ast.BinNumOpType.ADD, to_reuse=to_reuse)


def sub(left, right, to_reuse=False):
    return ast.BinNumExpr(left, right, ast.BinNumOpType.SUB, to_reuse=to_reuse)


def lt(left, right):
    return ast.CompExpr(left, right, ast.CompOpType.LT)


def lte(left, right):
    return ast.CompExpr(left, right, ast.CompOpType.LTE)


def gt(left, right):
    return ast.CompExpr(left, right, ast.CompOpType.GT)


def eq(left, right):
    return ast.CompExpr(left, right, ast.CompOpType.EQ)


BIN_EXPR_CLASSES = {
    (False, False): ast.BinNumExpr,
    (True, True): ast.BinVectorExpr,
    (True, False): ast.BinVectorNumExpr,
}


def apply_bin_op(left, right, op):
    """
    Finds binary expression class suitable for combination of left and right
    expressions depending on whether their output is scalar or vector and
    creates instance of this expression with specified operation.
    """
    exr_class = BIN_EXPR_CLASSES.get((left.output_size > 1, right.output_size > 1))
    if exr_class is None:
        # change the positions of left and right
        left, right = right, left
        exr_class = ast.BinVectorNumExpr

    return exr_class(left, right, op)


def apply_op_to_expressions(op, *exprs, to_reuse=False):
    if len(exprs) < 1:
        raise ValueError("At least one expression is required")
    if len(exprs) == 1:
        return exprs[0]

    def _inner(current_expr, *rest_exprs):
        if not rest_exprs:
            return current_expr

        return _inner(apply_bin_op(current_expr, rest_exprs[0], op), *rest_exprs[1:])

    result = _inner(apply_bin_op(exprs[0], exprs[1], op), *exprs[2:])
    result.to_reuse = to_reuse
    return result


def to_1d_array(var):
    return np.ravel(np.asarray(var))


def to_2d_array(var):
    shape = np.shape(var)
    if len(shape) == 2:
        x, y = shape
    else:
        x, y = 1, np.size(var)
    return np.reshape(np.asarray(var), (x, y))



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CodeTemplate, ImperativeCodeGenerator


class RubyCodeGenerator(ImperativeCodeGenerator):

    tpl_var_declaration = CodeTemplate("")
    tpl_num_value = CodeTemplate("{value}")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_return_statement = tpl_num_value
    tpl_array_index_access = CodeTemplate("{array_name}[{index}]")
    tpl_if_statement = CodeTemplate("if {if_def}")
    tpl_else_statement = CodeTemplate("else")
    tpl_block_termination = CodeTemplate("end")
    tpl_var_assignment = CodeTemplate("{var_name} = {value}")

    def add_function_def(self, name, args):
        func_def = f"def {name}({', '.join(args)})"
        self.add_code_line(func_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield
        self.add_block_termination()

    def method_invocation(self, method_name, obj, args):
        return f"({obj}).{method_name}({', '.join(map(str, args))})"

    def vector_init(self, values):
        return f"[{', '.join(values)}]"



import math
import sys
from collections import namedtuple

from m2cgen import ast
from m2cgen.interpreters.interpreter import BaseToCodeInterpreter


class BinExpressionDepthTrackingMixin(BaseToCodeInterpreter):
    """
    This mixin provides an ability to call a custom hook when depth of the
    binary expression reaches certain threshold.

    Subclasses must specify value for `bin_depth_threshold`.

    By default it creates a variable and assigns it the result of the incoming
    expression interpretation.

    Subclasses may override this default behaviour.
    """

    # disabled by default
    bin_depth_threshold = sys.maxsize

    def _pre_interpret_hook(self, expr, bin_depth=0, **kwargs):
        if not isinstance(expr, ast.BinExpr):
            return super()._pre_interpret_hook(expr, **kwargs)

        # We track depth of the binary expressions and call a hook if it
        # reaches specified threshold.
        if bin_depth >= self.bin_depth_threshold:
            return self.bin_depth_threshold_hook(expr, **kwargs), kwargs

        kwargs["bin_depth"] = bin_depth + 1
        return super()._pre_interpret_hook(expr, **kwargs)

    # Default implementation. Simply adds new variable.
    def bin_depth_threshold_hook(self, expr, **kwargs):
        if expr in self._cached_expr_results:
            return self._cached_expr_results[expr].var_name
        result = self._do_interpret(expr, **kwargs)
        var_name = self._cg.add_var_declaration(expr.output_size)
        self._cg.add_var_assignment(var_name, result, expr.output_size)
        return var_name


class LinearAlgebraMixin(BaseToCodeInterpreter):
    """
    This mixin provides simple way to interpret linear algebra expression as
    function invocation.

    It also provides flag `with_linear_algebra` which indicates whether
    linear algebra was used during interpretation. It can be used to add
    corresponding third party dependencies that provide linear algebra
    operations and/or data structures.
    """

    with_linear_algebra = False

    supported_bin_vector_ops = {}
    supported_bin_vector_num_ops = {}

    def interpret_bin_vector_expr(self, expr, extra_func_args=(), **kwargs):
        if expr.op not in self.supported_bin_vector_ops:
            raise NotImplementedError(f"Op '{expr.op.name}' is unsupported")

        self.with_linear_algebra = True

        function_name = self.supported_bin_vector_ops[expr.op]

        return self._cg.function_invocation(
            function_name,
            self._do_interpret(expr.left, **kwargs),
            self._do_interpret(expr.right, **kwargs),
            *extra_func_args)

    def interpret_bin_vector_num_expr(self, expr, extra_func_args=(), **kwargs):
        if expr.op not in self.supported_bin_vector_num_ops:
            raise NotImplementedError(f"Op '{expr.op.name}' is unsupported")

        self.with_linear_algebra = True

        function_name = self.supported_bin_vector_num_ops[expr.op]

        return self._cg.function_invocation(
            function_name,
            self._do_interpret(expr.left, **kwargs),
            self._do_interpret(expr.right, **kwargs),
            *extra_func_args)


Subroutine = namedtuple('Subroutine', ['name', 'expr'])


class SubroutinesMixin(BaseToCodeInterpreter):
    """
    This mixin provides ability to split the code into subroutines based on
    the size of the AST.

    Subclasses only need to implement `create_code_generator` method.

    Their code generators should implement 3 methods:
         - function_definition;
         - function_invocation;
         - add_return_statement.

    Interpreter should prepare at least one subroutine using method
    `enqueue_subroutine` and then call method `process_subroutine_queue` with
    instance of code generator, which will be populated with the result code.
    """

    # disabled by default
    ast_size_check_frequency = sys.maxsize
    ast_size_per_subroutine_threshold = sys.maxsize

    def __init__(self, *args, **kwargs):
        self._subroutine_idx = 0
        self.subroutine_expr_queue = []
        super().__init__(*args, **kwargs)

    def process_subroutine_queue(self, top_code_generator):
        """
        This method should be called from the interpreter to start processing
        subroutine queue.
        """
        self._subroutine_idx = 0

        while len(self.subroutine_expr_queue):
            self._reset_reused_expr_cache()
            subroutine = self.subroutine_expr_queue.pop(0)
            subroutine_code = self._process_subroutine(subroutine)
            top_code_generator.add_code_lines(subroutine_code)

    def enqueue_subroutine(self, name, expr):
        self.subroutine_expr_queue.append(Subroutine(name, expr))

    def _pre_interpret_hook(self, expr, ast_size_check_counter=0, **kwargs):
        if isinstance(expr, ast.BinExpr) and not expr.to_reuse:
            frequency = self._adjust_ast_check_frequency(expr)
            self.ast_size_check_frequency = min(frequency, self.ast_size_check_frequency)

            ast_size_check_counter += 1
            if ast_size_check_counter >= self.ast_size_check_frequency:
                ast_size_check_counter = 0
                ast_size = ast.count_exprs(expr)
                if ast_size > self.ast_size_per_subroutine_threshold:
                    function_name = self._get_subroutine_name()
                    self.enqueue_subroutine(function_name, expr)
                    return self._cg.function_invocation(function_name, self._feature_array_name), kwargs

            kwargs['ast_size_check_counter'] = ast_size_check_counter

        return super()._pre_interpret_hook(expr, **kwargs)

    def _adjust_ast_check_frequency(self, expr):
        """
        The logic below counts the number of non-binary expressions
        in a non-recursive branch of a binary expression to account
        for large tree-like models and adjust the size check frequency
        if necessary.
        """
        cnt = None
        if not isinstance(expr.left, ast.BinExpr):
            cnt = ast.count_exprs(expr.left, exclude_list={ast.BinExpr})
        elif not isinstance(expr.right, ast.BinExpr):
            cnt = ast.count_exprs(expr.right, exclude_list={ast.BinExpr})
        if cnt and cnt < self.ast_size_per_subroutine_threshold:
            return math.ceil(self.ast_size_per_subroutine_threshold / cnt)
        return self.ast_size_check_frequency

    def _process_subroutine(self, subroutine):
        """
        Handles single subroutine. Creates new code generator and defines a
        function for a given subroutine.
        """
        is_vector_output = subroutine.expr.output_size > 1

        self._cg = self.create_code_generator()

        with self._cg.function_definition(
                name=subroutine.name,
                args=[(True, self._feature_array_name)],
                is_vector_output=is_vector_output):
            last_result = self._do_interpret(subroutine.expr)
            self._cg.add_return_statement(last_result)

        return self._cg.finalize_and_get_generated_code()

    def _get_subroutine_name(self):
        subroutine_name = f"subroutine{self._subroutine_idx}"
        self._subroutine_idx += 1
        return subroutine_name

    # Methods to be implemented by subclasses.

    def create_code_generator(self):
        raise NotImplementedError


class PowExprInfixMixin(BaseToCodeInterpreter):
    """
    This mixin is used for languages that provide the exponentiation operation in a form
    of an infix operator. Such languages inlcude Haskell, F#, R and others.

    The operator used by default is "**", but it can be overriden by setting an appropriate
    value in `pow_operator`.
    """

    pow_operator = "**"

    infix_expressions = (*BaseToCodeInterpreter.infix_expressions, ast.PowExpr)

    def interpret_pow_expr(self, expr, left_precedence=None,
                           right_precedence=None, **kwargs):
        base_result = self._do_interpret(
            expr.base_expr, left_precedence=expr.precedence, **kwargs)
        exp_result = self._do_interpret(
            expr.exp_expr, right_precedence=expr.precedence, **kwargs)
        return self._cg.infix_expression(
            left=base_result,
            right=exp_result,
            op=self.pow_operator,
            wrap=self._wrap_infix_expr(expr, left_precedence, right_precedence))


class PowExprFunctionMixin(BaseToCodeInterpreter):
    """
    This mixin is used for languages that provide the exponentiation operation as a function.
    Such languages inlcude C, Java, Python and others.

    The name of the function must be set in `power_function_name` by implementing subclasses.
    """

    power_function_name = NotImplemented

    def interpret_pow_expr(self, expr, **kwargs):
        if self.power_function_name is NotImplemented:
            raise NotImplementedError("Power function is not provided")
        self.with_math_module = True
        base_result = self._do_interpret(expr.base_expr, **kwargs)
        exp_result = self._do_interpret(expr.exp_expr, **kwargs)
        return self._cg.function_invocation(self.power_function_name, base_result, exp_result)



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CLikeCodeGenerator


class CSharpCodeGenerator(CLikeCodeGenerator):

    scalar_type = "double"
    vector_type = "double[]"

    def add_class_def(self, class_name, modifier="public"):
        class_def = f"{modifier} static class {class_name} {{"
        self.add_code_line(class_def)
        self.increase_indent()

    def add_method_def(self, name, args, is_vector_output,
                       modifier="private"):
        return_type = self._get_var_declare_type(is_vector_output)
        func_args = ",".join([
            f"{self._get_var_declare_type(is_vector)} {n}"
            for is_vector, n in args])
        method_def = f"{modifier} static {return_type} {name}({func_args}) {{"
        self.add_code_line(method_def)
        self.increase_indent()

    def add_namespace_def(self, namespace):
        namespace_def = f"namespace {namespace} {{"
        self.add_code_line(namespace_def)
        self.increase_indent()

    @contextmanager
    def class_definition(self, class_name, modifier="public"):
        self.add_class_def(class_name, modifier=modifier)
        yield
        self.add_block_termination()

    @contextmanager
    def method_definition(self, name, args, is_vector_output,
                          modifier="private"):
        self.add_method_def(name, args, is_vector_output, modifier=modifier)
        yield
        self.add_block_termination()

    @contextmanager
    def namespace_definition(self, namespace):
        self.add_namespace_def(namespace)
        yield
        self.add_block_termination()

    def vector_init(self, values):
        return (f"new double[{len(values)}] {{{', '.join(values)}}}")

    def _get_var_declare_type(self, is_vector):
        return self.vector_type if is_vector else self.scalar_type

    def add_dependency(self, dep, modifier="static"):
        self.prepend_code_line(f"using {modifier} {dep};")



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CodeTemplate, FunctionalCodeGenerator


class ElixirCodeGenerator(FunctionalCodeGenerator):
    tpl_function_signature = CodeTemplate("{function_name} = fn ->")
    tpl_if_statement = CodeTemplate("cond do {if_def} -> ")
    tpl_else_statement = CodeTemplate("true ->")
    tpl_num_value = CodeTemplate("{value}")
    tpl_block_termination = CodeTemplate("end")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_module_definition = CodeTemplate("""defmodule {module_name} do
    @compile {{:inline, read: 2}}
    defp read(bin, pos) do
        <<_::size(pos)-unit(64)-binary, value::float, _::binary>> = bin
        value
    end
    defp list_to_binary(list) do
        for i <- list, into: <<>>, do: <<i::float>>
    end
    """)
    tpl_array_index_access = CodeTemplate("read({array_name},{index})")

    def add_if_termination(self):
        self.decrease_indent()
        self.add_code_line(self.tpl_block_termination())

    def get_func_name(self):
        func_name = f"func{self._func_idx}.()"
        self._func_idx += 1
        return func_name

    def add_function(self, function_name, function_body):
        self.add_code_line(self.tpl_function_signature(
            function_name=function_name.replace(".()", "")))
        self.increase_indent()
        self.add_code_lines(function_body)
        self.decrease_indent()
        self.add_code_line(self.tpl_block_termination())

    def add_function_def(self, name, args, is_scalar_output):

        func_args = ", ".join([n for _, n in args])
        function_def = f"def {name}({func_args}) do"
        self.add_code_line(function_def)

        self.increase_indent()
        self.add_code_line("input = list_to_binary(input)")

    def function_invocation(self, function_name, *args):
        function_args = ", ".join(args)
        return f"{function_name}({function_args})"

    @contextmanager
    def function_definition(self, name, args, is_scalar_output):
        self.increase_indent()
        self.add_function_def(name, args, is_scalar_output)
        yield
        self.decrease_indent()
        self.add_code_line(self.tpl_block_termination())
        self.decrease_indent()

    def vector_init(self, values):
        return f"[{', '.join(values)}]"



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CLikeCodeGenerator, CodeTemplate


class RustCodeGenerator(CLikeCodeGenerator):

    tpl_var_declaration = CodeTemplate("let {var_name}: {var_type};")
    tpl_num_value = CodeTemplate("{value}_f64")
    tpl_if_statement = CodeTemplate("if {if_def} {{")
    tpl_return_statement = CodeTemplate("{value}")

    scalar_type = "f64"
    vector_type = "Vec<f64>"

    def add_function_def(self, name, args, is_scalar_output):
        func_args = ", ".join([
            f"{n}: {self._get_var_declare_type(is_vector)}"
            for is_vector, n in args])
        return_type = self._get_var_declare_type(not is_scalar_output)
        function_def = f"fn {name}({func_args}) -> {return_type} {{"
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args, is_scalar_output):
        self.add_function_def(name, args, is_scalar_output)
        yield
        self.add_block_termination()

    def vector_init(self, values):
        return (f"vec![{', '.join(values)}]")

    def _get_var_declare_type(self, is_vector):
        return self.vector_type if is_vector else self.scalar_type



from enum import Enum
from inspect import getmembers, isclass
from sys import modules

import numpy as np


class Expr:
    output_size = 1

    # The precedence of this expression. Higher value means higher precedence.
    precedence = None

    # Setting this value to true serves as an indication that the result
    # of evaluation of this expression is being used in other expressions
    # and it's recommended to persist or cache it in some way.
    # The actual caching mechanism (if any) is left up to a specific
    # interpreter implementation to provide.
    to_reuse = False


class IdExpr(Expr):
    def __init__(self, expr, to_reuse=False):
        self.expr = expr
        self.to_reuse = to_reuse
        self.output_size = expr.output_size

    def __str__(self):
        return f"IdExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is IdExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class FeatureRef(Expr):
    def __init__(self, index):
        self.index = index

    def __str__(self):
        return f"FeatureRef({self.index})"

    def __eq__(self, other):
        return type(other) is FeatureRef and self.index == other.index

    def __hash__(self):
        return hash(self.index)


class BinExpr(Expr):
    pass


# Numeric Expressions.

class NumExpr(Expr):
    pass


class NumVal(NumExpr):
    def __init__(self, value, dtype=np.float64):
        self.value = dtype(value)

    def __str__(self):
        return f"NumVal({self.value})"

    def __eq__(self, other):
        return type(other) is NumVal and self.value == other.value

    def __hash__(self):
        return hash(self.value)


class AbsExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"AbsExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is AbsExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class AtanExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"AtanExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is AtanExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class ExpExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"ExpExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is ExpExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class LogExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"LogExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is LogExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class Log1pExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"Log1pExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is Log1pExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class SigmoidExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"SigmoidExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is SigmoidExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class SqrtExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"SqrtExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is SqrtExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class TanhExpr(NumExpr):
    def __init__(self, expr, to_reuse=False):
        assert expr.output_size == 1, "Only scalars are supported"

        self.expr = expr
        self.to_reuse = to_reuse

    def __str__(self):
        return f"TanhExpr({self.expr},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return type(other) is TanhExpr and self.expr == other.expr

    def __hash__(self):
        return hash(self.expr)


class PowExpr(NumExpr):
    precedence = 4

    def __init__(self, base_expr, exp_expr, to_reuse=False):
        assert base_expr.output_size == 1, "Only scalars are supported"
        assert exp_expr.output_size == 1, "Only scalars are supported"

        self.base_expr = base_expr
        self.exp_expr = exp_expr
        self.to_reuse = to_reuse

    def __str__(self):
        return (f"PowExpr({self.base_expr},{self.exp_expr},"
                f"to_reuse={self.to_reuse})")

    def __eq__(self, other):
        return (type(other) is PowExpr and
                self.base_expr == other.base_expr and
                self.exp_expr == other.exp_expr)

    def __hash__(self):
        return hash((self.base_expr, self.exp_expr))


class BinNumOpType(Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'


class BinNumExpr(NumExpr, BinExpr):
    def __init__(self, left, right, op, to_reuse=False):
        assert left.output_size == 1, "Only scalars are supported"
        assert right.output_size == 1, "Only scalars are supported"

        self.left = left
        self.right = right
        self.op = op
        self.to_reuse = to_reuse
        self.precedence = 3 if op in {BinNumOpType.MUL, BinNumOpType.DIV} else 2

    def __str__(self):
        return (f"BinNumExpr({self.left},{self.right},{self.op.name},"
                f"to_reuse={self.to_reuse})")

    def __eq__(self, other):
        return _eq_bin_exprs(self, other, type(self))

    def __hash__(self):
        return hash((self.left, self.right, self.op))


# Vector Expressions.

class VectorExpr(Expr):
    pass


class VectorVal(VectorExpr):

    def __init__(self, exprs):
        assert all(e.output_size == 1 for e in exprs), (
            "All expressions for VectorVal must be scalar")

        self.exprs = exprs
        self.output_size = len(exprs)

    def __str__(self):
        args = ",".join([str(e) for e in self.exprs])
        return f"VectorVal([{args}])"

    def __eq__(self, other):
        return (type(other) is VectorVal and
                self.output_size == other.output_size and
                all(i == j for i, j in zip(self.exprs, other.exprs)))

    def __hash__(self):
        return hash(tuple(self.exprs))


class SoftmaxExpr(VectorExpr):

    def __init__(self, exprs, to_reuse=False):
        assert all(e.output_size == 1 for e in exprs), (
            "All expressions for SoftmaxExpr must be scalar")

        self.exprs = exprs
        self.to_reuse = to_reuse
        self.output_size = len(exprs)

    def __str__(self):
        args = ",".join([str(e) for e in self.exprs])
        return f"SoftmaxExpr({args},to_reuse={self.to_reuse})"

    def __eq__(self, other):
        return (type(other) is SoftmaxExpr and
                self.output_size == other.output_size and
                all(i == j for i, j in zip(self.exprs, other.exprs)))

    def __hash__(self):
        return hash(tuple(self.exprs))


class BinVectorExpr(VectorExpr, BinExpr):

    def __init__(self, left, right, op):
        assert left.output_size > 1, "Only vectors are supported"
        assert left.output_size == right.output_size, (
            "Vectors must be of the same size")

        self.left = left
        self.right = right
        self.op = op
        self.output_size = left.output_size

    def __str__(self):
        return f"BinVectorExpr({self.left},{self.right},{self.op.name})"

    def __eq__(self, other):
        return _eq_bin_exprs(self, other, type(self))

    def __hash__(self):
        return hash((self.left, self.right, self.op))


class BinVectorNumExpr(VectorExpr, BinExpr):

    def __init__(self, left, right, op):
        assert left.output_size > 1, "Only vectors are supported"
        assert right.output_size == 1, "Only scalars are supported"

        self.left = left
        self.right = right
        self.op = op
        self.output_size = left.output_size

    def __str__(self):
        return f"BinVectorNumExpr({self.left},{self.right},{self.op.name})"

    def __eq__(self, other):
        return _eq_bin_exprs(self, other, type(self))

    def __hash__(self):
        return hash((self.left, self.right, self.op))


# Boolean Expressions.

class BoolExpr(Expr):
    pass


class CompOpType(Enum):
    GT = '>'
    GTE = '>='
    LT = '<'
    LTE = '<='
    EQ = '=='
    NOT_EQ = '!='

    @staticmethod
    def from_str_op(op):
        return COMP_OP_TYPE_MAPPING[op]


COMP_OP_TYPE_MAPPING = {e.value: e for e in CompOpType}


class CompExpr(BoolExpr):
    precedence = 1

    def __init__(self, left, right, op):
        assert left.output_size == 1, "Only scalars are supported"
        assert right.output_size == 1, "Only scalars are supported"

        self.left = left
        self.right = right
        self.op = op

    def __str__(self):
        return f"CompExpr({self.left},{self.right},{self.op.name})"

    def __eq__(self, other):
        return _eq_bin_exprs(self, other, type(self))

    def __hash__(self):
        return hash((self.left, self.right, self.op))


# Control Expressions.

class CtrlExpr(Expr):
    size = None


class IfExpr(CtrlExpr):
    def __init__(self, test, body, orelse):
        assert body.output_size == orelse.output_size, (
            "body and orelse expressions should have the same output size")

        self.test = test
        self.body = body
        self.orelse = orelse
        self.output_size = body.output_size

    def __str__(self):
        return f"IfExpr({self.test},{self.body},{self.orelse})"

    def __eq__(self, other):
        return (type(other) is IfExpr and
                self.test == other.test and
                self.body == other.body and
                self.orelse == other.orelse)

    def __hash__(self):
        return hash((self.test, self.body, self.orelse))


TOTAL_NUMBER_OF_EXPRESSIONS = len(getmembers(modules[__name__], isclass))


NESTED_EXPRS_MAPPINGS = [
    ((BinExpr, CompExpr), lambda e: [e.left, e.right]),
    ((PowExpr), lambda e: [e.base_expr, e.exp_expr]),
    ((VectorVal, SoftmaxExpr), lambda e: e.exprs),
    ((IfExpr), lambda e: [e.test, e.body, e.orelse]),
    ((AbsExpr, AtanExpr, ExpExpr, IdExpr, LogExpr, Log1pExpr,
      SigmoidExpr, SqrtExpr, TanhExpr),
     lambda e: [e.expr]),
]


def count_exprs(expr, exclude_list=None):
    expr_type = type(expr)
    excluded = tuple(exclude_list) if exclude_list else ()

    init = 1
    if issubclass(expr_type, excluded):
        init = 0

    if isinstance(expr, (NumVal, FeatureRef)):
        return init

    for tpes, nested_f in NESTED_EXPRS_MAPPINGS:
        if issubclass(expr_type, tpes):
            return init + sum(map(
                lambda e: count_exprs(e, exclude_list),
                nested_f(expr)))

    expr_type_name = expr_type.__name__
    raise ValueError(f"Unexpected expression type '{expr_type_name}'")


def _eq_bin_exprs(expr_one, expr_two, expected_type):
    return (type(expr_one) is expected_type and
            type(expr_two) is expected_type and
            expr_one.left == expr_two.left and
            expr_one.right == expr_two.right and
            expr_one.op == expr_two.op)



from contextlib import contextmanager

from m2cgen.ast import CompOpType
from m2cgen.interpreters.code_generator import CodeTemplate, FunctionalCodeGenerator


class FSharpCodeGenerator(FunctionalCodeGenerator):
    tpl_function_signature = CodeTemplate("let {function_name} =")
    tpl_if_statement = CodeTemplate("if {if_def} then")
    tpl_else_statement = CodeTemplate("else")
    tpl_num_value = CodeTemplate("{value}")
    tpl_infix_expression = CodeTemplate("{left} {op} {right}")
    tpl_array_index_access = CodeTemplate("{array_name}.[{index}]")

    def add_if_termination(self):
        self.decrease_indent()

    def add_function_def(self, name, args):
        func_args = " ".join(
            [f"({n} : double{' list' if is_vector else ''})"
             for is_vector, n in args])
        function_def = f"let {name} {func_args} ="
        self.add_code_line(function_def)
        self.increase_indent()

    @contextmanager
    def function_definition(self, name, args):
        self.add_function_def(name, args)
        yield
        self.decrease_indent()

    def vector_init(self, values):
        return f"[{'; '.join(values)}]"

    def _comp_op_overwrite(self, op):
        if op == CompOpType.EQ:
            return "="
        elif op == CompOpType.NOT_EQ:
            return "<>"
        else:
            return op.value



from pathlib import Path

from m2cgen.ast import BinNumOpType
from m2cgen.interpreters.interpreter import ImperativeToCodeInterpreter
from m2cgen.interpreters.java.code_generator import JavaCodeGenerator
from m2cgen.interpreters.mixins import LinearAlgebraMixin, PowExprFunctionMixin, SubroutinesMixin
from m2cgen.interpreters.utils import get_file_content


class JavaInterpreter(ImperativeToCodeInterpreter,
                      PowExprFunctionMixin,
                      LinearAlgebraMixin,
                      SubroutinesMixin):

    # The below numbers have been determined experimentally and are subject
    # to adjustments in future.
    ast_size_check_frequency = 100
    ast_size_per_subroutine_threshold = 4600

    supported_bin_vector_ops = {
        BinNumOpType.ADD: "addVectors",
    }

    supported_bin_vector_num_ops = {
        BinNumOpType.MUL: "mulVectorNumber",
    }

    abs_function_name = "Math.abs"
    atan_function_name = "Math.atan"
    exponent_function_name = "Math.exp"
    logarithm_function_name = "Math.log"
    log1p_function_name = "Math.log1p"
    power_function_name = "Math.pow"
    sigmoid_function_name = "sigmoid"
    softmax_function_name = "softmax"
    sqrt_function_name = "Math.sqrt"
    tanh_function_name = "Math.tanh"

    with_sigmoid_expr = False
    with_softmax_expr = False

    def __init__(self, package_name=None, class_name="Model", indent=4,
                 function_name="score", *args, **kwargs):
        self.package_name = package_name
        self.class_name = class_name
        self.indent = indent
        self.function_name = function_name

        # We don't provide any code generator as for each subroutine we will
        # create a new one and concatenate their results into top_cg created
        # in .interpret() method.
        super().__init__(None, *args, **kwargs)

    def interpret(self, expr):
        top_cg = self.create_code_generator()

        if self.package_name:
            top_cg.add_package_name(self.package_name)

        with top_cg.class_definition(self.class_name):

            # Since we use SubroutinesMixin, we already have logic
            # of adding methods. We create first subroutine for incoming
            # expression and call `process_subroutine_queue` method.
            self.enqueue_subroutine(self.function_name, expr)
            self.process_subroutine_queue(top_cg)

            current_dir = Path(__file__).absolute().parent

            if self.with_linear_algebra:
                filename = current_dir / "linear_algebra.java"
                top_cg.add_code_lines(get_file_content(filename))

            if self.with_softmax_expr:
                filename = current_dir / "softmax.java"
                top_cg.add_code_lines(get_file_content(filename))

            if self.with_sigmoid_expr:
                filename = current_dir / "sigmoid.java"
                top_cg.add_code_lines(get_file_content(filename))

        return top_cg.finalize_and_get_generated_code()

    # Required by SubroutinesMixin to create new code generator for
    # each subroutine.
    def create_code_generator(self):
        return JavaCodeGenerator(indent=self.indent)

    def interpret_softmax_expr(self, expr, **kwargs):
        self.with_softmax_expr = True
        return super().interpret_softmax_expr(expr, **kwargs)

    def interpret_sigmoid_expr(self, expr, **kwargs):
        self.with_sigmoid_expr = True
        return super().interpret_sigmoid_expr(expr, **kwargs)



from contextlib import contextmanager

from m2cgen.interpreters.code_generator import CLikeCodeGenerator


class JavaCodeGenerator(CLikeCodeGenerator):

    scalar_output_type = "double"
    vector_output_type = "double[]"

    def add_class_def(self, class_name, modifier="public"):
        class_def = f"{modifier} class {class_name} {{"
        self.add_code_line(class_def)
        self.increase_indent()

    def add_method_def(self, name, args, is_vector_output, modifier="public"):
        return_type = self._get_var_declare_type(is_vector_output)

        func_args = ",".join([
            f"{self._get_var_declare_type(is_vector)} {n}"
            for is_vector, n in args])
        method_def = f"{modifier} static {return_type} {name}({func_args}) {{"
        self.add_code_line(method_def)
        self.increase_indent()

    def add_package_name(self, package_name):
        self.add_code_line(f"package {package_name};")

    @contextmanager
    def class_definition(self, class_name):
        self.add_class_def(class_name)
        yield
        self.add_block_termination()

    @contextmanager
    def method_definition(self, name, args, is_vector_output,
                          modifier="public"):
        self.add_method_def(name, args, is_vector_output, modifier=modifier)
        yield
        self.add_block_termination()

    def vector_init(self, values):
        return f"new {self.vector_output_type} {{{', '.join(values)}}}"

    def _get_var_declare_type(self, is_vector):
        return self.vector_output_type if is_vector else self.scalar_output_type

    # Method `function_definition` is required by SubroutinesMixin.
    # We already have this functionality in `method_definition` method.
    function_definition = method_definition
